import asyncio

import pytest

import python_utils
from python_utils import types


@pytest.mark.asyncio
async def test_abatcher() -> None:
    async for batch in python_utils.abatcher(python_utils.acount(stop=9), 3):
        assert len(batch) == 3

    async for batch in python_utils.abatcher(python_utils.acount(stop=2), 3):
        assert len(batch) == 2


@pytest.mark.asyncio
async def test_abatcher_timed() -> None:
    batches: types.List[types.List[int]] = []
    async for batch in python_utils.abatcher(
        python_utils.acount(stop=10, delay=0.08), interval=0.1
    ):
        batches.append(batch)

    assert batches == [[0, 1, 2], [3, 4], [5, 6], [7, 8], [9]]
    assert len(batches) == 5


@pytest.mark.asyncio
async def test_abatcher_timed_with_timeout() -> None:
    async def generator() -> types.AsyncIterator[int]:
        # Test if the timeout is respected
        yield 0
        yield 1
        await asyncio.sleep(0.11)

        # Test if the timeout is respected
        yield 2
        yield 3
        await asyncio.sleep(0.11)

        # Test if exceptions are handled correctly
        await asyncio.wait_for(asyncio.sleep(1), timeout=0.05)

        # Test if StopAsyncIteration is handled correctly
        yield 4

    batcher = python_utils.abatcher(generator(), interval=0.1)
    assert await batcher.__anext__() == [0, 1]
    assert await batcher.__anext__() == [2, 3]

    with pytest.raises(asyncio.TimeoutError):
        await batcher.__anext__()

    with pytest.raises(StopAsyncIteration):
        await batcher.__anext__()


def test_batcher() -> None:
    batch = []
    for batch in python_utils.batcher(range(9), 3):
        assert len(batch) == 3

    for batch in python_utils.batcher(range(4), 3):
        assert batch is not None

    assert len(batch) == 1



import asyncio

import pytest

from python_utils import types
from python_utils.aio import acontainer, acount, adict


@pytest.mark.asyncio
async def test_acount(monkeypatch: pytest.MonkeyPatch) -> None:
    sleeps: types.List[float] = []

    async def mock_sleep(delay: float) -> None:
        sleeps.append(delay)

    monkeypatch.setattr(asyncio, 'sleep', mock_sleep)

    async for _i in acount(delay=1, stop=3.5):
        pass

    assert len(sleeps) == 4
    assert sum(sleeps) == 4


@pytest.mark.asyncio
async def test_acontainer() -> None:
    async def async_gen() -> types.AsyncIterable[int]:
        yield 1
        yield 2
        yield 3

    async def empty_gen() -> types.AsyncIterable[int]:
        if False:
            yield 1

    assert await acontainer(async_gen) == [1, 2, 3]
    assert await acontainer(async_gen()) == [1, 2, 3]
    assert await acontainer(async_gen, set) == {1, 2, 3}
    assert await acontainer(async_gen(), set) == {1, 2, 3}
    assert await acontainer(async_gen, list) == [1, 2, 3]
    assert await acontainer(async_gen(), list) == [1, 2, 3]
    assert await acontainer(async_gen, tuple) == (1, 2, 3)
    assert await acontainer(async_gen(), tuple) == (1, 2, 3)
    assert await acontainer(empty_gen) == []
    assert await acontainer(empty_gen()) == []
    assert await acontainer(empty_gen, set) == set()
    assert await acontainer(empty_gen(), set) == set()
    assert await acontainer(empty_gen, list) == list()
    assert await acontainer(empty_gen(), list) == list()
    assert await acontainer(empty_gen, tuple) == tuple()
    assert await acontainer(empty_gen(), tuple) == tuple()


@pytest.mark.asyncio
async def test_adict() -> None:
    async def async_gen() -> types.AsyncIterable[types.Tuple[int, int]]:
        yield 1, 2
        yield 3, 4
        yield 5, 6

    async def empty_gen() -> types.AsyncIterable[types.Tuple[int, int]]:
        if False:
            yield 1, 2

    assert await adict(async_gen) == {1: 2, 3: 4, 5: 6}
    assert await adict(async_gen()) == {1: 2, 3: 4, 5: 6}
    assert await adict(empty_gen) == {}
    assert await adict(empty_gen()) == {}



import typing
from unittest.mock import MagicMock

import pytest

from python_utils.decorators import sample, wraps_classmethod

T = typing.TypeVar('T')


@pytest.fixture
def random(monkeypatch: pytest.MonkeyPatch) -> MagicMock:
    mock = MagicMock()
    monkeypatch.setattr(
        'python_utils.decorators.random.random', mock, raising=True
    )
    return mock


def test_sample_called(random: MagicMock) -> None:
    demo_function = MagicMock()
    decorated = sample(0.5)(demo_function)
    random.return_value = 0.4
    decorated()
    random.return_value = 0.0
    decorated()
    args = [1, 2]
    kwargs = {'1': 1, '2': 2}
    decorated(*args, **kwargs)
    demo_function.assert_called_with(*args, **kwargs)
    assert demo_function.call_count == 3


def test_sample_not_called(random: MagicMock) -> None:
    demo_function = MagicMock()
    decorated = sample(0.5)(demo_function)
    random.return_value = 0.5
    decorated()
    random.return_value = 1.0
    decorated()
    assert demo_function.call_count == 0


class SomeClass:
    @classmethod
    def some_classmethod(cls, arg: T) -> T:
        return arg

    @classmethod
    def some_annotated_classmethod(cls, arg: int) -> int:
        return arg


def test_wraps_classmethod() -> None:
    some_class = SomeClass()
    some_class.some_classmethod = MagicMock()  # type: ignore[method-assign]
    wrapped_method = wraps_classmethod(SomeClass.some_classmethod)(
        some_class.some_classmethod
    )
    wrapped_method(123)
    some_class.some_classmethod.assert_called_with(123)


def test_wraps_annotated_classmethod() -> None:
    some_class = SomeClass()
    some_class.some_annotated_classmethod = MagicMock()  # type: ignore[method-assign]
    wrapped_method = wraps_classmethod(SomeClass.some_annotated_classmethod)(
        some_class.some_annotated_classmethod
    )
    wrapped_method(123)
    some_class.some_annotated_classmethod.assert_called_with(123)



# mypy: disable-error-code=misc
import pytest

from python_utils.loguru import Logurud

loguru = pytest.importorskip('loguru')


def test_logurud() -> None:
    class MyClass(Logurud):
        pass

    my_class = MyClass()
    my_class.debug('debug')
    my_class.info('info')
    my_class.warning('warning')
    my_class.error('error')
    my_class.critical('critical')
    my_class.exception('exception')
    my_class.log(0, 'log')



from python_utils import import_, types


def test_import_globals_relative_import() -> None:
    for i in range(-1, 5):
        relative_import(i)


def relative_import(level: int) -> None:
    locals_: types.Dict[str, types.Any] = {}
    globals_ = {'__name__': 'python_utils.import_'}
    import_.import_global('.formatters', locals_=locals_, globals_=globals_)
    assert 'camel_to_underscore' in globals_


def test_import_globals_without_inspection() -> None:
    locals_: types.Dict[str, types.Any] = {}
    globals_: types.Dict[str, types.Any] = {'__name__': __name__}
    import_.import_global(
        'python_utils.formatters', locals_=locals_, globals_=globals_
    )
    assert 'camel_to_underscore' in globals_


def test_import_globals_single_method() -> None:
    locals_: types.Dict[str, types.Any] = {}
    globals_: types.Dict[str, types.Any] = {'__name__': __name__}
    import_.import_global(
        'python_utils.formatters',
        ['camel_to_underscore'],
        locals_=locals_,
        globals_=globals_,
    )
    assert 'camel_to_underscore' in globals_


def test_import_globals_with_inspection() -> None:
    import_.import_global('python_utils.formatters')
    assert 'camel_to_underscore' in globals()


def test_import_globals_missing_module() -> None:
    import_.import_global(
        'python_utils.spam', exceptions=ImportError, locals_=locals()
    )
    assert 'camel_to_underscore' in globals()


def test_import_locals_missing_module() -> None:
    import_.import_global(
        'python_utils.spam', exceptions=ImportError, globals_=globals()
    )
    assert 'camel_to_underscore' in globals()



"""
Setup script for the python-utils package.

This script uses setuptools to package the python-utils library. It reads
metadata from the `python_utils/__about__.py` file and the `README.rst` file to
populate the package information. The script also defines the package
requirements and optional dependencies for different use cases such as logging,
documentation, and testing.
"""

import pathlib

import setuptools

# pyright: reportUnknownMemberType=false

# To prevent importing about and thereby breaking the coverage info we use this
# exec hack
about: dict[str, str] = {}
with open('python_utils/__about__.py') as fp:
    exec(fp.read(), about)

_readme_path = pathlib.Path(__file__).parent / 'README.rst'
if _readme_path.exists() and _readme_path.is_file():
    long_description = _readme_path.read_text()
else:
    long_description = 'See http://pypi.python.org/pypi/python-utils/'

if __name__ == '__main__':
    setuptools.setup(
        python_requires='>=3.9.0',
        name='python-utils',
        version=about['__version__'],
        author=about['__author__'],
        author_email=about['__author_email__'],
        description=about['__description__'],
        url=about['__url__'],
        license='BSD',
        packages=setuptools.find_packages(
            exclude=['_python_utils_tests', '*.__pycache__'],
        ),
        package_data={'python_utils': ['py.typed']},
        long_description=long_description,
        install_requires=['typing_extensions>3.10.0.2'],
        extras_require={
            'loguru': [
                'loguru',
            ],
            'docs': [
                'mock',
                'sphinx',
                'python-utils',
            ],
            'tests': [
                'ruff',
                'pyright',
                'pytest',
                'pytest-cov',
                'pytest-mypy',
                'pytest-asyncio',
                'sphinx',
                'types-setuptools',
                'loguru',
                'loguru-mypy',
                'mypy-ipython',
                'blessings',
            ],
        },
        classifiers=['License :: OSI Approved :: BSD License'],
    )
